﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.IsolatedStorage;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Resources;
using System.Windows.Shapes;
using System.Xml.Linq;

namespace XapLoadingAndCaching
{
  /// <summary>
  /// Proof of concept of an XAP download and cache manager.
  /// </summary>
  public static class SimpleXapManager
  {
    /// <summary>
    /// Load the specified XAP file with the specified version number either from cache
    /// or download it first.
    /// </summary>
    /// <remarks>
    /// This code is very critical for the application that uses it. Add error handling and
    /// check contingent to ensure that it will not fail.
    /// </remarks>
    public static void LoadXap(string name, string version)
    {
      // Do not load XAP file more than once.
      if (_loadedXaps.ContainsKey(name))
        return;

      using (var store = IsolatedStorageFile.GetUserStoreForSite())
      {
        string cachedFileName = GetCachedFileName(name, version);
        if (store.FileExists(cachedFileName))
        {
          // Take file from cache.
          IsolatedStorageFileStream fileStream = store.OpenFile(cachedFileName, FileMode.Open, FileAccess.Read);
          LoadXap(fileStream);
          _loadedXaps.Add(name, null);
        }
        else
        {
          // Clear older version of file, if any exists.
          ClearCache(store, name);

          // Download file from site of origin.
          var webClient = new WebClient();
          webClient.OpenReadAsync(new Uri(name, UriKind.Relative));
          webClient.OpenReadCompleted += (sender, e) =>
          {
            using (var store2 = IsolatedStorageFile.GetUserStoreForSite())
            {
              // Save copy of new file in store.
              int length = (int)e.Result.Length;
              byte[] buffer = new byte[length];
              e.Result.Read(buffer, 0, length);
              using (var fileStream = store2.CreateFile(cachedFileName))
              {
                fileStream.Write(buffer, 0, length);
              }
              LoadXap(e.Result);
              _loadedXaps.Add(name, null);
            }
          }; // Info: It is valid to set the completed handler after starting the async operation.
        }
      }
    }

    /// <summary>
    /// Determines whether the specified XAP file is already loaded in the current app-domain.
    /// </summary>
    public static bool IsXapLoaded(string name)
    {
      return _loadedXaps.ContainsKey(name);
    }

    /// <summary>
    /// Empty the cache.
    /// </summary>
    public static void ClearCache()
    {
      using (var store = IsolatedStorageFile.GetUserStoreForSite())
      {
        // HACK: Just delete everything.
        string[] files = store.GetFileNames("*.*");
        foreach (var file in files)
        {
          try
          {
            store.DeleteFile(file);
          }
          catch (Exception ex)
          {
            // Happens when file is in use by current application.
            Debug.WriteLine(ex.Message);
          }
        }
      }
    }

    /// <summary>
    /// Load XAP file from a stream into the current app-domain.
    /// </summary>
    static void LoadXap(Stream stream)
    {
      StreamResourceInfo sri = new StreamResourceInfo(stream, null);

      // Get a list of all assembly part names from application manifest.
      string appManifest = new StreamReader(Application.GetResourceStream(sri, new Uri("AppManifest.xaml", UriKind.Relative)).Stream).ReadToEnd();
      XElement deploy = XDocument.Parse(appManifest).Root;
      Debug.Assert(deploy != null);
      List<XElement> parts = (from assemblyParts in deploy.Elements().Elements() select assemblyParts).ToList();

      // Load all assembly parts.
      foreach (var source in parts.Select(element => element.Attribute("Source")))
      {
        var streamInfo = Application.GetResourceStream(sri, new Uri(source.Value, UriKind.Relative));
        new AssemblyPart().Load(streamInfo.Stream);
      }
    }

    /// <summary>
    /// Clear old XAP files from storage.
    /// </summary>
    static void ClearCache(IsolatedStorageFile store, string name)
    {
      string[] files = store.GetFileNames(name + "-*");
      foreach (var file in files)
      {
        try
        {
          store.DeleteFile(file);
        }
        catch (Exception ex)
        {
          Debug.Assert(false, "Cannot delete old XAP file from cache: " + ex.Message);
        }
      }
    }

    static string GetCachedFileName(string name, string version)
    {
      _loadedXaps.Remove(name);
      if (!String.IsNullOrEmpty(version))
        name += "-" + version;
      return name;
    }

    static readonly Dictionary<string, object> _loadedXaps = new Dictionary<string, object>();
  }
}